@@ -30,6 +30,7 @@
 #include <errno.h>
 
 #include <bluetooth/bluetooth.h>
+#include <bluetooth/sdp.h>
 
 #include <glib.h>
 #include <dbus/dbus.h>
@@ -37,12 +38,14 @@
 
 #include "logging.h"
 
-#include "avdtp.h"
 #include "device.h"
+#include "avdtp.h"
 #include "a2dp.h"
 #include "error.h"
 #include "sink.h"
 #include "dbus-common.h"
+#include "../src/adapter.h"
+#include "../src/device.h"
 
 #define STREAM_SETUP_RETRY_TIMER 2
 
@@ -53,24 +56,126 @@ struct pending_request {
 };
 
 struct sink {
+	struct audio_device *dev;
 	struct avdtp *session;
 	struct avdtp_stream *stream;
 	unsigned int cb_id;
-	uint8_t state;
+	guint dc_id;
+	avdtp_session_state_t session_state;
+	avdtp_state_t stream_state;
+	sink_state_t state;
 	struct pending_request *connect;
 	struct pending_request *disconnect;
 	DBusConnection *conn;
 };
 
-static void pending_request_free(struct pending_request *pending)
+struct sink_state_callback {
+	sink_state_cb cb;
+	void *user_data;
+	unsigned int id;
+};
+
+static GSList *sink_callbacks = NULL;
+
+static unsigned int avdtp_callback_id = 0;
+
+static const char *state2str(sink_state_t state)
+{
+	switch (state) {
+	case SINK_STATE_DISCONNECTED:
+		return "disconnected";
+	case SINK_STATE_CONNECTING:
+		return "connecting";
+	case SINK_STATE_CONNECTED:
+		return "connected";
+	case SINK_STATE_PLAYING:
+		return "playing";
+	default:
+		error("Invalid sink state %d", state);
+		return NULL;
+	}
+}
+
+static void sink_set_state(struct audio_device *dev, sink_state_t new_state)
+{
+	struct sink *sink = dev->sink;
+	const char *state_str;
+	sink_state_t old_state = sink->state;
+	GSList *l;
+
+	sink->state = new_state;
+
+	state_str = state2str(new_state);
+	if (state_str)
+		emit_property_changed(dev->conn, dev->path,
+					AUDIO_SINK_INTERFACE, "State",
+					DBUS_TYPE_STRING, &state_str);
+
+	for (l = sink_callbacks; l != NULL; l = l->next) {
+		struct sink_state_callback *cb = l->data;
+		cb->cb(dev, old_state, new_state, cb->user_data);
+	}
+}
+
+static void avdtp_state_callback(struct audio_device *dev,
+					struct avdtp *session,
+					avdtp_session_state_t old_state,
+					avdtp_session_state_t new_state,
+					void *user_data)
+{
+	struct sink *sink = dev->sink;
+
+	if (sink == NULL)
+		return;
+
+	switch (new_state) {
+	case AVDTP_SESSION_STATE_DISCONNECTED:
+		if (sink->state != SINK_STATE_CONNECTING) {
+			gboolean value = FALSE;
+			g_dbus_emit_signal(dev->conn, dev->path,
+					AUDIO_SINK_INTERFACE, "Disconnected",
+					DBUS_TYPE_INVALID);
+			emit_property_changed(dev->conn, dev->path,
+					AUDIO_SINK_INTERFACE, "Connected",
+					DBUS_TYPE_BOOLEAN, &value);
+			device_remove_disconnect_watch(dev->btd_dev,
+							sink->dc_id);
+		}
+		sink_set_state(dev, SINK_STATE_DISCONNECTED);
+		break;
+	case AVDTP_SESSION_STATE_CONNECTING:
+		sink_set_state(dev, SINK_STATE_CONNECTING);
+		break;
+	case AVDTP_SESSION_STATE_CONNECTED:
+		break;
+	}
+
+	sink->session_state = new_state;
+}
+
+static void pending_request_free(struct audio_device *dev,
+					struct pending_request *pending)
 {
 	if (pending->conn)
 		dbus_connection_unref(pending->conn);
 	if (pending->msg)
 		dbus_message_unref(pending->msg);
+	if (pending->id)
+		a2dp_source_cancel(dev, pending->id);
+
 	g_free(pending);
 }
 
+static void disconnect_cb(struct btd_device *btd_dev, void *user_data)
+{
+	struct audio_device *device = user_data;
+	struct sink *sink = device->sink;
+
+	debug("Sink: disconnect %s", device->path);
+
+	avdtp_close(sink->session, sink->stream);
+}
+
 static void stream_state_changed(struct avdtp_stream *stream,
 					avdtp_state_t old_state,
 					avdtp_state_t new_state,
@@ -86,12 +191,6 @@ static void stream_state_changed(struct avdtp_stream *stream,
 
 	switch (new_state) {
 	case AVDTP_STATE_IDLE:
-		value = FALSE;
-		g_dbus_emit_signal(dev->conn, dev->path, AUDIO_SINK_INTERFACE,
-					"Disconnected", DBUS_TYPE_INVALID);
-		emit_property_changed(dev->conn, dev->path,
-					AUDIO_SINK_INTERFACE, "Connected",
-					DBUS_TYPE_BOOLEAN, &value);
 		if (sink->disconnect) {
 			DBusMessage *reply;
 			struct pending_request *p;
@@ -101,7 +200,7 @@ static void stream_state_changed(struct avdtp_stream *stream,
 
 			reply = dbus_message_new_method_return(p->msg);
 			g_dbus_send_message(p->conn, reply);
-			pending_request_free(p);
+			pending_request_free(dev, p);
 		}
 
 		if (sink->session) {
@@ -112,7 +211,8 @@ static void stream_state_changed(struct avdtp_stream *stream,
 		sink->cb_id = 0;
 		break;
 	case AVDTP_STATE_OPEN:
-		if (old_state == AVDTP_STATE_CONFIGURED) {
+		if (old_state == AVDTP_STATE_CONFIGURED &&
+				sink->state == SINK_STATE_CONNECTING) {
 			value = TRUE;
 			g_dbus_emit_signal(dev->conn, dev->path,
 						AUDIO_SINK_INTERFACE,
@@ -122,6 +222,9 @@ static void stream_state_changed(struct avdtp_stream *stream,
 						AUDIO_SINK_INTERFACE,
 						"Connected",
 						DBUS_TYPE_BOOLEAN, &value);
+			sink->dc_id = device_add_disconnect_watch(dev->btd_dev,
+								disconnect_cb,
+								dev, NULL);
 		} else if (old_state == AVDTP_STATE_STREAMING) {
 			value = FALSE;
 			g_dbus_emit_signal(dev->conn, dev->path,
@@ -133,6 +236,7 @@ static void stream_state_changed(struct avdtp_stream *stream,
 						"Playing",
 						DBUS_TYPE_BOOLEAN, &value);
 		}
+		sink_set_state(dev, SINK_STATE_CONNECTED);
 		break;
 	case AVDTP_STATE_STREAMING:
 		value = TRUE;
@@ -141,6 +245,7 @@ static void stream_state_changed(struct avdtp_stream *stream,
 		emit_property_changed(dev->conn, dev->path,
 					AUDIO_SINK_INTERFACE, "Playing",
 					DBUS_TYPE_BOOLEAN, &value);
+		sink_set_state(dev, SINK_STATE_PLAYING);
 		break;
 	case AVDTP_STATE_CONFIGURED:
 	case AVDTP_STATE_CLOSING:
@@ -149,7 +254,7 @@ static void stream_state_changed(struct avdtp_stream *stream,
 		break;
 	}
 
-	sink->state = new_state;
+	sink->stream_state = new_state;
 }
 
 static DBusHandlerResult error_failed(DBusConnection *conn,
@@ -163,7 +268,7 @@ static gboolean stream_setup_retry(gpointer user_data)
 	struct sink *sink = user_data;
 	struct pending_request *pending = sink->connect;
 
-	if (sink->state >= AVDTP_STATE_OPEN) {
+	if (sink->stream_state >= AVDTP_STATE_OPEN) {
 		debug("Stream successfully created, after XCASE connect:connect");
 		if (pending->msg) {
 			DBusMessage *reply;
@@ -177,7 +282,7 @@ static gboolean stream_setup_retry(gpointer user_data)
 	}
 
 	sink->connect = NULL;
-	pending_request_free(pending);
+	pending_request_free(sink->dev, pending);
 
 	return FALSE;
 }
@@ -191,6 +296,8 @@ static void stream_setup_complete(struct avdtp *session, struct a2dp_sep *sep,
 
 	pending = sink->connect;
 
+	pending->id = 0;
+
 	if (stream) {
 		debug("Stream successfully created");
 
@@ -201,7 +308,7 @@ static void stream_setup_complete(struct avdtp *session, struct a2dp_sep *sep,
 		}
 
 		sink->connect = NULL;
-		pending_request_free(pending);
+		pending_request_free(sink->dev, pending);
 
 		return;
 	}
@@ -217,7 +324,7 @@ static void stream_setup_complete(struct avdtp *session, struct a2dp_sep *sep,
 		if (pending->msg)
 			error_failed(pending->conn, pending->msg, "Stream setup failed");
 		sink->connect = NULL;
-		pending_request_free(pending);
+		pending_request_free(sink->dev, pending);
 		debug("Stream setup failed : %s", avdtp_strerror(err));
 	}
 }
@@ -365,6 +472,7 @@ static void discovery_complete(struct avdtp *session, GSList *seps, struct avdtp
 	struct pending_request *pending;
 	struct avdtp_local_sep *lsep;
 	struct avdtp_remote_sep *rsep;
+	struct a2dp_sep *sep;
 	GSList *caps = NULL;
 	int id;
 
@@ -396,7 +504,13 @@ static void discovery_complete(struct avdtp *session, GSList *seps, struct avdtp
 		goto failed;
 	}
 
-	id = a2dp_source_config(sink->session, stream_setup_complete,
+	sep = a2dp_source_get(session, rsep);
+	if (!sep) {
+		error("Unable to get a local source SEP");
+		goto failed;
+	}
+
+	id = a2dp_source_config(sink->session, sep, stream_setup_complete,
 				caps, sink);
 	if (id == 0)
 		goto failed;
@@ -407,7 +521,7 @@ static void discovery_complete(struct avdtp *session, GSList *seps, struct avdtp
 failed:
 	if (pending->msg)
 		error_failed(pending->conn, pending->msg, "Stream setup failed");
-	pending_request_free(pending);
+	pending_request_free(sink->dev, pending);
 	sink->connect = NULL;
 	avdtp_unref(sink->session);
 	sink->session = NULL;
@@ -421,6 +535,11 @@ gboolean sink_setup_stream(struct sink *sink, struct avdtp *session)
 	if (session && !sink->session)
 		sink->session = avdtp_ref(session);
 
+	if (!sink->session)
+		return FALSE;
+
+	avdtp_set_auto_disconnect(sink->session, FALSE);
+
 	if (avdtp_discover(sink->session, discovery_complete, sink) < 0)
 		return FALSE;
 
@@ -447,17 +566,17 @@ static DBusMessage *sink_connect(DBusConnection *conn,
 		return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
 						"%s", strerror(EBUSY));
 
-	if (sink->state >= AVDTP_STATE_OPEN)
+	if (sink->stream_state >= AVDTP_STATE_OPEN)
 		return g_dbus_create_error(msg, ERROR_INTERFACE
 						".AlreadyConnected",
 						"Device Already Connected");
 
-	avdtp_set_auto_disconnect(sink->session, FALSE);
-
 	if (!sink_setup_stream(sink, NULL))
 		return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
 						"Failed to create a stream");
 
+	dev->auto_connect = FALSE;
+
 	pending = sink->connect;
 
 	pending->conn = dbus_connection_ref(conn);
@@ -485,7 +604,7 @@ static DBusMessage *sink_disconnect(DBusConnection *conn,
 		return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
 						"%s", strerror(EBUSY));
 
-	if (sink->state < AVDTP_STATE_OPEN) {
+	if (sink->stream_state < AVDTP_STATE_OPEN) {
 		DBusMessage *reply = dbus_message_new_method_return(msg);
 		if (!reply)
 			return NULL;
@@ -520,7 +639,7 @@ static DBusMessage *sink_is_connected(DBusConnection *conn,
 	if (!reply)
 		return NULL;
 
-	connected = (sink->state >= AVDTP_STATE_CONFIGURED);
+	connected = (sink->stream_state >= AVDTP_STATE_CONFIGURED);
 
 	dbus_message_append_args(reply, DBUS_TYPE_BOOLEAN, &connected,
 					DBUS_TYPE_INVALID);
@@ -532,9 +651,11 @@ static DBusMessage *sink_get_properties(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
 	struct audio_device *device = data;
+	struct sink *sink = device->sink;
 	DBusMessage *reply;
 	DBusMessageIter iter;
 	DBusMessageIter dict;
+	const char *state;
 	gboolean value;
 
 	reply = dbus_message_new_method_return(msg);
@@ -549,13 +670,18 @@ static DBusMessage *sink_get_properties(DBusConnection *conn,
 			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
 
 	/* Playing */
-	value = (device->sink->state == AVDTP_STATE_STREAMING);
+	value = (sink->stream_state == AVDTP_STATE_STREAMING);
 	dict_append_entry(&dict, "Playing", DBUS_TYPE_BOOLEAN, &value);
 
 	/* Connected */
-	value = (device->sink->state >= AVDTP_STATE_CONFIGURED);
+	value = (sink->stream_state >= AVDTP_STATE_CONFIGURED);
 	dict_append_entry(&dict, "Connected", DBUS_TYPE_BOOLEAN, &value);
 
+	/* State */
+	state = state2str(sink->state);
+	if (state)
+		dict_append_entry(&dict, "State", DBUS_TYPE_STRING, &state);
+
 	dbus_message_iter_close_container(&iter, &dict);
 
 	return reply;
@@ -593,10 +719,10 @@ static void sink_free(struct audio_device *dev)
 		avdtp_unref(sink->session);
 
 	if (sink->connect)
-		pending_request_free(sink->connect);
+		pending_request_free(dev, sink->connect);
 
 	if (sink->disconnect)
-		pending_request_free(sink->disconnect);
+		pending_request_free(dev, sink->disconnect);
 
 	g_free(sink);
 	dev->sink = NULL;
@@ -606,7 +732,7 @@ static void path_unregister(void *data)
 {
 	struct audio_device *dev = data;
 
-	info("Unregistered interface %s on path %s",
+	debug("Unregistered interface %s on path %s",
 		AUDIO_SINK_INTERFACE, dev->path);
 
 	sink_free(dev);
@@ -620,16 +746,26 @@ void sink_unregister(struct audio_device *dev)
 
 struct sink *sink_init(struct audio_device *dev)
 {
+	struct sink *sink;
+
 	if (!g_dbus_register_interface(dev->conn, dev->path,
 					AUDIO_SINK_INTERFACE,
 					sink_methods, sink_signals, NULL,
 					dev, path_unregister))
 		return NULL;
 
-	info("Registered interface %s on path %s",
+	debug("Registered interface %s on path %s",
 		AUDIO_SINK_INTERFACE, dev->path);
 
-	return g_new0(struct sink, 1);
+	if (avdtp_callback_id == 0)
+		avdtp_callback_id = avdtp_add_state_cb(avdtp_state_callback,
+									NULL);
+
+	sink = g_new0(struct sink, 1);
+
+	sink->dev = dev;
+
+	return sink;
 }
 
 gboolean sink_is_active(struct audio_device *dev)
@@ -646,7 +782,7 @@ avdtp_state_t sink_get_state(struct audio_device *dev)
 {
 	struct sink *sink = dev->sink;
 
-	return sink->state;
+	return sink->stream_state;
 }
 
 gboolean sink_new_stream(struct audio_device *dev, struct avdtp *session,
@@ -667,3 +803,45 @@ gboolean sink_new_stream(struct audio_device *dev, struct avdtp *session,
 
 	return TRUE;
 }
+
+gboolean sink_shutdown(struct sink *sink)
+{
+	if (!sink->stream)
+		return FALSE;
+
+	if (avdtp_close(sink->session, sink->stream) < 0)
+		return FALSE;
+
+	return TRUE;
+}
+
+unsigned int sink_add_state_cb(sink_state_cb cb, void *user_data)
+{
+	struct sink_state_callback *state_cb;
+	static unsigned int id = 0;
+
+	state_cb = g_new(struct sink_state_callback, 1);
+	state_cb->cb = cb;
+	state_cb->user_data = user_data;
+	state_cb->id = ++id;
+
+	sink_callbacks = g_slist_append(sink_callbacks, state_cb);
+
+	return state_cb->id;
+}
+
+gboolean sink_remove_state_cb(unsigned int id)
+{
+	GSList *l;
+
+	for (l = sink_callbacks; l != NULL; l = l->next) {
+		struct sink_state_callback *cb = l->data;
+		if (cb && cb->id == id) {
+			sink_callbacks = g_slist_remove(sink_callbacks, cb);
+			g_free(cb);
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
