@@ -39,4 +39,5 @@ static void storage_exit(void)
 {
 }
 
-BLUETOOTH_PLUGIN_DEFINE("storage", VERSION, storage_init, storage_exit)
+BLUETOOTH_PLUGIN_DEFINE(storage, VERSION,
+			BLUETOOTH_PLUGIN_PRIORITY_DEFAULT, storage_init, storage_exit)
