@@ -29,6 +29,7 @@ void manager_cleanup(DBusConnection *conn, const char *path);
 
 const char *manager_get_base_path(void);
 struct btd_adapter *manager_find_adapter(const bdaddr_t *sba);
+struct btd_adapter *manager_find_adapter_by_address(const char *address);
 struct btd_adapter *manager_find_adapter_by_path(const char *path);
 struct btd_adapter *manager_find_adapter_by_id(int id);
 GSList *manager_get_adapters(void);
@@ -36,6 +37,7 @@ int manager_register_adapter(int id, gboolean devup);
 int manager_unregister_adapter(int id);
 int manager_start_adapter(int id);
 int manager_stop_adapter(int id);
+void manager_add_adapter(const char *path);
 int manager_get_default_adapter();
 void manager_set_default_adapter(int id);
 int manager_update_adapter(uint16_t id, uint8_t svc, gboolean starting);
